from pathlib import Path

import yaml
from dynaconf import LazySettings

from . import envvar_helper, mixin_helper

try:
    from cally.idp.defaults import DEFAULTS as IDP_DEFAULTS  # type: ignore
except ModuleNotFoundError:
    IDP_DEFAULTS: dict = {}  # type: ignore[no-redef]


def load(obj: LazySettings, *args, **kwargs) -> None:  # noqa: ARG001
    """
    Load a cally yaml file, with a resolution order of defaults, environment
    defaults, service, then environment variables.

    cally.yml
    ```yaml
    defaults:
      providers:
        example:
          foo: bar
    dev:
      defaults:
        providers:
          random:
            alias: cats
      services:
        example-service:
          stack_vars:
            foo: bar
    """
    config_file = Path(obj.settings_file_for_dynaconf)
    loaded = {}
    if config_file.exists():
        loaded = yaml.safe_load(config_file.read_text())

    # Defaults
    obj.update(IDP_DEFAULTS)
    obj.update(loaded.get('defaults', {}))
    if obj.cally_env is not None:
        obj.update(environment=obj.cally_env)
        obj.update(loaded.get(obj.cally_env, {}).get('defaults', {}))

    if obj.cally_env:
        services = loaded.get(obj.cally_env, {}).get('services', {})
        for service, config in services.items():
            obj.update({'services': {service: config}})
            service_obj = obj.get('services', {}).get(service, {})
            mixin_helper(service_obj, loaded, config)

    # Process Env Vars
    envvar_helper(obj)



import click

from ..tools.provider import ProviderBuilder


@click.group()
def provider() -> None:
    pass


@click.command()
@click.option('--source', default='hashicorp')
@click.option('--provider', required=True)
@click.option('--version', required=True)
def build(source: str, provider: str, version: str):
    click.secho(f'Generating {provider} ({version}) provider')
    ProviderBuilder(source=source, provider=provider, version=version).build()


provider.add_command(build)



import json
import os
from inspect import getfile
from pathlib import Path
from tempfile import TemporaryDirectory
from unittest import TestCase, mock

from cally.cdk import CallyStack
from cally.cli.config.service import CallyServiceConfig
from cally.cli.config.terraform_service import CallyStackServiceConfig

from .constants import HOME_ENVS
from .exceptions import CallyTestingTestdataError


class CallyTestHarness(TestCase):
    """Cally provides a Test Harness as good base for building robust unit testing
    for your IDP and IAC. It takes care of ensuring a clean environment, and discrete
    working directory for each unittest, along with providing some useful methods for
    getting testdata or default service objects.
    """

    working: TemporaryDirectory
    _testdata: Path

    def setUp(self):
        self.current_working = Path().cwd()
        self.working = TemporaryDirectory()
        self.env_patcher = mock.patch.dict(
            os.environ,
            {
                HOME_ENVS.get(os.name, 'UNKNOWN'): self.working.name,
                "LC_ALL": os.environ.get('LC_ALL', 'C.UTF-8'),
                "LANG": os.environ.get('LANG', 'C.UTF-8'),
            },
            clear=True,
        )
        self.env_patcher.start()
        os.chdir(self.working.name)

    def tearDown(self):
        self.env_patcher.stop()
        os.chdir(self.current_working)
        self.working.cleanup()

    @property
    def testdata(self) -> Path:
        """Attempts to find a directory called 'testdata' within your tests directory"""
        if getattr(self, '_testdata', None) is None:
            for parent in Path(getfile(self.__class__)).parents:
                testdata = Path(parent, 'testdata')
                if testdata.exists() and testdata.is_dir():
                    self._testdata = testdata
                    return self._testdata
                if parent.is_dir() and parent.name == 'tests':
                    break
            raise CallyTestingTestdataError(
                "No testdata found, expected structure <project_root>/tests/testdata"
            )
        return self._testdata

    @staticmethod
    def get_cally_config(service='test', environment='cally') -> CallyServiceConfig:
        """
        Returns a CallyServiceConfig object, with a service and environment
        pre-configured as 'test' and 'cally'
        """
        config = CallyServiceConfig()
        config.config_file = Path('not-required.yaml')
        config.service = service
        config.environment = environment
        return config

    @staticmethod
    def get_cally_stack_config(
        service='test', environment='cally', stack_type='CallyStack'
    ) -> CallyStackServiceConfig:
        """
        Returns a CallyStackServiceConfig object, with a service and environment
        pre-configured as service='test', environment='cally', stack_type='CallyStack'
        """
        config = CallyStackServiceConfig()
        config.config_file = Path('not-required.yaml')
        config.service = service
        config.environment = environment
        config.settings.stack_type = stack_type
        return config

    def get_test_file(self, filename) -> Path:
        """Loads a the specified file path as a Path object, from the testdata
        directory
        """
        return Path(self.testdata, filename)

    def load_test_file(self, filename) -> str:
        """Loads a testdata file and reads it in as text"""
        return self.get_test_file(filename).read_text()

    def load_json_file(self, filename) -> dict:
        """Loads a json testdata file and returns it as a dictionary"""
        return json.loads(self.load_test_file(filename))


class CallyTfTestHarness(CallyTestHarness):
    """This is an extension of the Cally Test harness, specifically for testing
    Stacks."""

    def synth_stack(self, stack: CallyStack) -> dict:
        """When passed a CallyStack object, synths it and returns the dictionary
        payload generated by the cdktf. Can be used to unit test your stacks.
        """
        stack.synth_stack(self.working.name)
        output_file = Path(self.working.name, 'stacks', stack.name, 'cdk.tf.json')
        return json.loads(output_file.read_text())



import click

from . import CallyCommandClass, CallyConfig, config_types, mixins
from .options import CALLY_SERVICE_OPTIONS


class CallyServiceConfig(CallyConfig[config_types.CallyService], mixins.CallyService):
    CALLY_TYPE = config_types.CallyService

    @property
    def _settings_kwargs(self) -> dict:
        return {
            'loaders': ['cally.cli.config.loaders.service'],
            'cally_env': self.environment,
            'cally_service': self.service,
        }


class CallyServiceConfigContext(click.Context):
    def __init__(self, *args, **kwargs) -> None:
        if kwargs.get('obj', None) is None:
            kwargs.update(obj=CallyServiceConfig())
        super().__init__(*args, **kwargs)


def CallyServiceCommand():  # noqa: N802
    class CallyServiceCommandClass(CallyCommandClass):
        context_class = CallyServiceConfigContext
        default_cally_options = CALLY_SERVICE_OPTIONS

    return CallyServiceCommandClass



import inspect
from copy import deepcopy
from dataclasses import dataclass, field, make_dataclass
from importlib import import_module
from types import MappingProxyType
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Tuple, Union

from cdktf import (
    App,
    TerraformBackend,
    TerraformOutput,
    TerraformProvider,
    TerraformResource,
    TerraformStack,
)
from constructs import Construct

if TYPE_CHECKING:
    from cally.cli.config.config_types import CallyStackService


@dataclass
class CallyResourceAttributes:
    _instantiated_attributes: dict = field(default_factory=dict)

    def instantiate_attributes(self) -> None:
        attrs = {}
        for key, val in self.__dict__.items():
            if key == '_instantiated_attributes':
                continue
            if hasattr(val, '_cdktf_resource'):
                attrs.update({key: val.construct_resource()})
                continue
            if isinstance(val, list):
                attrs.update(
                    {
                        key: [
                            (
                                x.construct_resource()
                                if hasattr(x, '_cdktf_resource')
                                else x
                            )
                            for x in val
                        ]
                    }
                )
                continue
            attrs.update({key: val})
        self._instantiated_attributes = attrs

    def to_dict(self) -> dict:
        if not self._instantiated_attributes:
            self.instantiate_attributes()
        return self._instantiated_attributes


class CallyResource:
    """This resource is for referencing the underlying CDK for Terraform resource,
    so that it can be later instantiated. This allows us to maniuplate variables on
    ``__init__``, along with setting defaults. Top level resources *must* be wrapped in
    this class, and attributes with a class specification *may* be wrapped with this
    for the purposes of setting defaults. Otherwise as long as the object structure
    is valid for the resource, the CDK will do the RightThings :TM:, however they will
    *not* be type checked and an output may succeed, but fail to validate/plan/apply.

    :param tf_identifier: This is required for the top level resource class, and
                          invalid for resource attributes. It is recommended to
                          specify this as the first argument rather than as a keyword.
                          This is expected to be a string, for consistency it should
                          follow the Identifier requirements set out by Terraform, but
                          the CDKTF will attempt to normalise it.

    """

    _cdktf_resource: Any  # This is probably a callable TerraformResource
    _tf_identifier: Optional[str]
    _instantiated_resource: TerraformResource
    attributes: CallyResourceAttributes

    #: Provider name, this is so cally knows where to load the underlying cdktf resource
    #: from. For example ``random``.
    provider: str
    #: Resource name, that cally will load the matching class name from, for example
    #: RandomPet, lives in ``random_pet``.
    resource: str
    #: A dictionary of the default values to set on instantion, if not passed through
    #: as kwargs. Wrapping in a MappingProxyType avoids the mutable warnings. For
    #: example ``MappingProxyType({'prefix': 'foo'})``
    defaults: Union[dict, MappingProxyType]

    def __init__(self, tf_identifier: Optional[str] = None, **kwargs) -> None:
        module = import_module(f'cally.providers.{self.provider}.{self.resource}')
        self._cdktf_resource = getattr(module, self.__class__.__name__)
        self.attributes = self._build_attributes(tf_identifier, **kwargs)

    def __str__(self) -> str:
        if self.tf_identifier:
            return f'${{{self.tf_resource}.{self.tf_identifier}.id}}'
        return self.__class__.__name__

    def __getattr__(self, item: str) -> Optional[str]:
        # TODO: This likely could use some improvement
        if item.startswith('__jsii'):
            return getattr(self._instantiated_resource, item)
        if item in {'attributes', 'defaults', '_instantiated_resource'}:
            return None
        if self.tf_identifier:
            return f'${{{self.tf_resource}.{self.tf_identifier}.{item}}}'
        return None

    def _get_attribute_default(self, name: str) -> Any:
        if self.defaults is None:
            return None
        return deepcopy(self.defaults.get(name, None))

    def _build_attributes(
        self, tf_identifier: Optional[str] = None, **kwargs
    ) -> CallyResourceAttributes:
        func = self._cdktf_resource.__init__  # type: ignore
        parameters = inspect.signature(func).parameters
        fields = [
            (name, param.annotation, self._get_attribute_default(name))
            for name, param in parameters.items()
            if param.annotation is not inspect._empty and name not in {'scope'}
        ]
        name = f'{self.__class__.__name__}CallyAttributes'
        cls = make_dataclass(name, fields, bases=(CallyResourceAttributes,))
        if tf_identifier:
            # Some newer provider releases appear to use 'id_'
            id_field = 'id'
            if 'id_' in parameters:
                id_field = 'id_'
            kwargs.update({id_field: tf_identifier})
        self._tf_identifier = tf_identifier
        return cls(**kwargs)

    @property
    def tf_identifier(self) -> Optional[str]:
        return self._tf_identifier

    @property
    def tf_resource(self) -> Optional[str]:
        if self.resource.startswith('data_'):
            return f'data.{self.resource[5:]}'
        return self.resource

    def construct_resource(
        self,
        scope: Optional[Construct] = None,
        provider: Optional[TerraformProvider] = None,
    ) -> TerraformResource:
        if getattr(self, '_instantiated_resource', None) is None:
            attrubtes = self.attributes.to_dict()
            if scope and provider:
                attrubtes.update(
                    {
                        'scope': scope,
                        'provider': provider,
                    }
                )
            self._instantiated_resource = self._cdktf_resource(**attrubtes)
        return self._instantiated_resource


class CallyStack:
    """This forms the parent of your custom stacks. It has a number of convenience
    methods, and it is intended that you override ``__init__``, call super, and
    construct your own stack from there.
    """

    _outputs: List[Tuple[str, str]]
    _providers: Dict[str, TerraformProvider]
    _resources: List[CallyResource]
    service: 'CallyStackService'

    def __init__(self, service: 'CallyStackService') -> None:
        self.service = service

    def add_output(self, tf_identifier: str, output: str) -> None:
        """Adds a terraform output to the resulting generated terraform."""
        self.outputs.append((tf_identifier, output))

    def add_resource(self, resource: CallyResource) -> None:
        """Adds a CallyResource to the stack, all resources must be added to the
        stack from them to be including in the resutling output
        """
        self.resources.append(resource)

    def add_resources(self, resources: List[CallyResource]) -> None:
        """Adds a list of CallyResources to the stack"""
        self.resources.extend(resources)

    def get_backend(self) -> TerraformBackend:
        mod = import_module('cdktf')
        return getattr(mod, self.service.backend_type)

    def get_provider(self, scope: Construct, provider: str) -> TerraformProvider:
        if provider not in self.providers:
            # google_beta -> GoogleBetaProvider
            resource = f"{provider.capitalize().replace('_', '')}Provider"
            module = import_module(f'cally.providers.{provider}.provider')
            cls = getattr(module, resource)
            self.providers.update(
                {
                    provider: cls(
                        scope, id=provider, **self.service.providers.get(provider, {})
                    )
                }
            )
        prov = self.providers.get(provider)
        if prov is None:
            raise ValueError("Provider instantion failed")
        return prov

    @property
    def name(self) -> str:
        return self.service.name

    @property
    def environment(self) -> str:
        return self.service.environment

    @property
    def resources(self) -> List[CallyResource]:
        if getattr(self, '_resources', None) is None:
            self._resources = []
        return self._resources

    @property
    def outputs(self) -> List[Tuple[str, str]]:
        if getattr(self, '_outputs', None) is None:
            self._outputs = []
        return self._outputs

    @property
    def providers(self) -> Dict[str, TerraformProvider]:
        if getattr(self, '_providers', None) is None:
            self._providers = {}
        return self._providers

    def synth_stack(self, outdir='cdktf.out'):
        stack = self

        class MyStack(TerraformStack):

            def __init__(self, scope: Construct) -> None:
                super().__init__(scope, stack.name)
                for resource in stack.resources:
                    resource.construct_resource(
                        self,
                        provider=stack.get_provider(self, resource.provider),
                    )
                for tf_identifier, value in stack.outputs:
                    TerraformOutput(self, tf_identifier, value=value)
                stack.get_backend()(self, **stack.service.backend_config)

        app = App(outdir=outdir)
        MyStack(app)

        app.synth()



import os
import shutil
import subprocess
from pathlib import Path
from string import Template
from tempfile import TemporaryDirectory

from ..constants import PROVIDER_CDKTF, PROVIDER_PYPROJECT


class ProviderBuilder:
    provider: str
    source: str
    version: str

    def __init__(self, source: str, provider: str, version: str) -> None:
        self.source = source
        self.provider = provider
        self.version = version

    @property
    def title(self) -> str:
        return self.provider.title().replace('_', '')

    @property
    def provider_path(self) -> str:
        return f"{self.source}/{self.provider.replace('_', '-')}"

    def build(self, output: Path = Path('build')) -> str:
        build_output = Path(output, self.provider)
        build_path = Path(build_output, 'cally', 'providers')
        build_path.mkdir(parents=True, exist_ok=True)
        cdktf_output = build_path.absolute().as_posix()

        cwd = Path.cwd()
        with TemporaryDirectory() as tmpdir:
            os.chdir(tmpdir)
            Path('cdktf.json').write_text(
                Template(PROVIDER_CDKTF).substitute(
                    provider=self.provider_path,
                    version=self.version,
                    path=cdktf_output,
                )
            )
            cdktf_command = str(shutil.which('cdktf'))

            result = subprocess.run(
                f'{cdktf_command} get', shell=True, check=False, stdout=subprocess.PIPE
            )
            message = f"Generated {self.provider} provider"
            if result.returncode != 0:
                message = f"Failed generating {self.provider} provider"
            command_output = result.stdout.decode('utf8').strip()
            os.chdir(cwd)

        Path(build_output, 'pyproject.toml').write_text(
            Template(PROVIDER_PYPROJECT).substitute(
                title=self.title, version=self.version
            )
        )
        return f'{message}\n\n{command_output}'



import click
import yaml

from ..config.environment import CallyEnvironmentCommand, CallyEnvironmentConfig
from ..config.service import CallyServiceCommand, CallyServiceConfig


@click.group()
def config() -> None:
    pass


@click.command(cls=CallyServiceCommand())
@click.pass_obj
def print_service(config: CallyServiceConfig):
    """Prints the service config as YAML"""
    click.secho(yaml.dump(config.settings.to_dict()))


@click.command(cls=CallyEnvironmentCommand())
@click.pass_obj
def print_services(config: CallyEnvironmentConfig):
    """Prints all the services in an environment config as YAML"""
    click.secho(yaml.dump(config.settings.to_dict()))


config.add_command(print_service)
config.add_command(print_services)



import os
from pathlib import Path
from tempfile import TemporaryDirectory
from types import TracebackType
from typing import Type

from cally.cdk import stacks

from ..config.config_types import CallyStackService


class Action:
    _cwd: Path
    _tmp_dir: TemporaryDirectory
    service: CallyStackService

    def __init__(self, service: CallyStackService) -> None:
        self.service = service

    def __enter__(self) -> 'Action':
        self._tmp_dir = TemporaryDirectory()
        self._cwd = Path().cwd()
        return self

    def __exit__(
        self,
        exc_type: Type[BaseException],
        exc_value: BaseException,
        traceback: TracebackType,
    ) -> None:
        os.chdir(self._cwd)
        self._tmp_dir.cleanup()

    @property
    def tmp_dir(self) -> str:
        return self._tmp_dir.name

    @property
    def output_path(self) -> Path:
        return Path(self.tmp_dir, 'stacks', self.service.name)

    @property
    def output_file(self) -> Path:
        return Path(self.output_path, 'cdk.tf.json')

    def synth_stack(
        self,
    ) -> None:
        # TODO: fix typing here
        cls = getattr(stacks, self.service.stack_type)
        cls(self.service).synth_stack(self.tmp_dir)

    def print(self) -> str:
        self.synth_stack()
        return self.output_file.read_text()



from dynaconf import LazySettings


def mixin_helper(obj: LazySettings, loaded: dict, service: dict) -> None:
    mixins = service.pop('mixins', [])
    # Resolution Order: Global Mixins, Env Mixins, Service
    # The last to be loaded wins.
    for mixin in [x.strip() for x in mixins if len(x.strip()) > 0]:
        obj.update(loaded.get('mixins', {}).get(mixin, {}))
        obj.update(loaded.get(obj.cally_env, {}).get('mixins', {}).get(mixin, {}))
    obj.update(service)


def envvar_helper(obj: LazySettings) -> None:
    # Environment Variables
    prefix: str = obj.envvar_prefix_for_dynaconf
    remove = len(prefix) + 1  # PREFIX_ lenth
    for key, val in obj.environ.items():
        if not key.startswith(prefix):
            continue
        if key in {f'{prefix}_SERVICE', f'{prefix}_ENVIRONMENT'}:
            continue
        obj.update({key[remove:].lower(): val})

    # Clear out Service/Environment
    obj.unset(f'{prefix}_ENV')
    obj.unset(f'{prefix}_SERVICE')



from pathlib import Path
from typing import Callable, ClassVar, Generic, List, Optional, TypeVar

import click
from dynaconf import Dynaconf

from . import config_types
from .options import CALLY_CONFIG_OPTIONS
from .validators import BASE_CALLY_CONFIG

T = TypeVar(
    "T",
    config_types.CallyEnvironment,
    config_types.CallyProject,
    config_types.CallyService,
    config_types.CallyStackService,
)


class CallyConfig(Generic[T]):
    CALLY_TYPE: ClassVar[Callable]
    _config_file: Path
    _config: T
    _settings: Dynaconf

    @property
    def config_file(self) -> Optional[Path]:
        return self._config_file

    @config_file.setter
    def config_file(self, value: Path):
        self._config_file = value

    @property
    def _settings_kwargs(self) -> dict:
        return {'loaders': []}

    @property
    def settings(self):
        if getattr(self, '_settings', None) is None:
            self._settings = Dynaconf(
                environments=False,
                envvar_prefix='CALLY',
                root_path=Path().cwd(),
                settings_file=self.config_file,
                merge_enabled=True,
                core_loaders=[],
                validators=BASE_CALLY_CONFIG,
                **self._settings_kwargs,
            )
        return self._settings

    @property
    def config(
        self,
    ) -> T:
        if getattr(self, '_config', None) is None:
            self._config = self.CALLY_TYPE(
                **config_types.filter_dataclass_props(self.settings, self.CALLY_TYPE)
            )
        return self._config


class CallyConfigContext(click.Context):
    def __init__(self, *args, **kwargs) -> None:
        if kwargs.get('obj', None) is None:
            kwargs.update(obj=CallyConfig())
        super().__init__(*args, **kwargs)


class CallyCommandClass(click.Command):
    context_class = CallyConfigContext
    default_cally_options: List[click.Option] = CALLY_CONFIG_OPTIONS

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.params.extend(self.default_cally_options)


def CallyCommand():  # noqa: N802
    return CallyCommandClass



from contextlib import suppress
from importlib import import_module
from pathlib import Path
from pkgutil import walk_packages
from typing import List

import click
from click import Group

from ._version import VERSION as __version__  # noqa: N811
from .config import CallyConfig


# TODO: Investigate the lazy loader as this might become quite slow
# https://click.palletsprojects.com/en/8.1.x/complex/#defining-the-lazy-group
def get_commands(package_name: str) -> List:
    """
    Convenience method for collecting all available commands
    """
    groups = {}
    package = import_module(package_name)
    for _, name, _is_pkg in walk_packages(package.__path__):
        full_name = f'{package.__name__}.{name}'
        module = import_module(full_name)
        for sub_module in dir(module):
            attr = getattr(module, sub_module)
            if isinstance(attr, Group):
                groups.update({name: attr})

    return [val for (_, val) in groups.items() if isinstance(val, click.core.Command)]


@click.group()
@click.version_option(__version__)
def cally() -> None:
    """
    Top level click command group for Cally
    """


commands = get_commands('cally.cli.commands')
with suppress(ModuleNotFoundError):
    commands.extend(get_commands('cally.idp.commands'))
for command in commands:
    cally.add_command(command)



from functools import update_wrapper

import click

from . import CallyCommandClass, CallyConfig, config_types, mixins
from .options import CALLY_SERVICE_OPTIONS


class CallyStackServiceConfig(
    CallyConfig[config_types.CallyStackService], mixins.CallyService
):
    CALLY_TYPE = config_types.CallyStackService

    @property
    def _settings_kwargs(self) -> dict:
        return {
            'loaders': ['cally.cli.config.loaders.service'],
            'cally_env': self.environment,
            'cally_service': self.service,
        }


class CallyStackServiceConfigContext(click.Context):
    def __init__(self, *args, **kwargs) -> None:
        if kwargs.get('obj', None) is None:
            kwargs.update(obj=CallyStackServiceConfig())
        super().__init__(*args, **kwargs)


def CallyStackServiceCommand():  # noqa: N802
    class CallyStackServiceCommandClass(CallyCommandClass):
        context_class = CallyStackServiceConfigContext
        default_cally_options = CALLY_SERVICE_OPTIONS

    return CallyStackServiceCommandClass



from pathlib import Path

import click

from ..config.terraform_service import CallyStackServiceCommand, CallyStackServiceConfig
from ..tools import terraform


@click.group()
def tf() -> None:
    pass


@click.command(name='print', cls=CallyStackServiceCommand())
@click.pass_obj
def print_template(config: CallyStackServiceConfig):
    with terraform.Action(service=config.config) as action:
        click.secho(action.print())


@click.command(name='write', cls=CallyStackServiceCommand())
@click.option(
    '--output',
    type=click.Path(path_type=Path),
    default=Path(Path.cwd(), 'cdk.tf.json'),
    help='Output path for the terraform json',
)
@click.pass_obj
def write_template(config: CallyStackServiceConfig, output: Path):
    with terraform.Action(service=config.config) as action:
        output.write_text(action.print())
        click.secho(f'Template written to {output}')


tf.add_command(print_template)
tf.add_command(write_template)



from typing import Optional


class CallyEnvironment:
    _environment: Optional[str] = None

    @property
    def environment(self) -> Optional[str]:
        return self._environment

    @environment.setter
    def environment(self, value: str):
        self._environment = value


class CallyService(CallyEnvironment):
    _service: Optional[str]

    @property
    def service(self) -> Optional[str]:
        return self._service

    @service.setter
    def service(self, value: str):
        self._service = value



import importlib.util
import sys
from pathlib import Path

from cally.testing import CallyTestHarness
from cally.testing.exceptions import CallyTestingTestdataError


class CallyTestingTests(CallyTestHarness):

    def test_testdata_raises(self):
        testpath = Path('tests')
        testpath.mkdir()
        self.get_test_file('testing/no_testdata.py').read_text()
        output = Path(testpath, 'no_testdata.py')
        output.write_text(self.get_test_file('testing/no_testdata.py').read_text())
        spec = importlib.util.spec_from_file_location("no_testdata", output)
        no_testdata = importlib.util.module_from_spec(spec)
        sys.modules["no_testdata"] = no_testdata
        spec.loader.exec_module(no_testdata)
        with self.assertRaises(CallyTestingTestdataError):
            _ = no_testdata.NoTestData().testdata



from dynaconf import Validator  # type: ignore


def is_lowercase(value: str) -> bool:
    return value.islower()


OPERATIONS_MESSAGE = {
    'operations': '{name} must {operation} {op_value}, but is: {value}'
}

BASE_CALLY_CONFIG = [
    Validator(
        'NAME',
        is_type_of=str,
        condition=is_lowercase,
        messages={'condition': 'Name must be lowercase'},
    ),
    Validator(
        'ENVIRONMENT',
        is_type_of=str,
        messages=OPERATIONS_MESSAGE,
    ),
    Validator(
        'STACK_TYPE',
        is_type_of=str,
        messages=OPERATIONS_MESSAGE,
    ),
    Validator('PROVIDERS', is_type_of=dict, messages=OPERATIONS_MESSAGE),
    Validator(
        'STACK_VARS',
        is_type_of=dict,
        messages=OPERATIONS_MESSAGE,
    ),
    Validator("BACKEND.type", is_type_of=str, messages=OPERATIONS_MESSAGE),
    Validator("BACKEND.path", is_type_of=str, messages=OPERATIONS_MESSAGE),
    Validator("BACKEND.key", is_type_of=str, messages=OPERATIONS_MESSAGE),
]



from contextlib import suppress
from importlib import import_module
from inspect import isclass
from pkgutil import walk_packages

from .. import CallyStack


def gather_stacks(package_name: str) -> None:
    package = import_module(package_name)
    for _, name, is_pkg in walk_packages(package.__path__):  # noqa: B007
        module = import_module(f'{package.__name__}.{name}')
        for sub_module in dir(module):
            if sub_module == 'CallyStack':
                continue
            attr = getattr(module, sub_module)
            if not isclass(attr):
                continue
            if issubclass(attr, CallyStack):
                globals()[sub_module] = attr


gather_stacks(__package__)
with suppress(ModuleNotFoundError):
    gather_stacks('cally.idp.stacks')



from dataclasses import asdict, dataclass, field, fields, is_dataclass
from typing import Any, Optional

from ..exceptions import ObjectNotDataclassError


def filter_dataclass_props(data: dict, cls: Any) -> dict:
    if not is_dataclass(cls):
        raise ObjectNotDataclassError(f'{cls} is not of type dataclass')
    return {x.name: data.get(x.name) for x in fields(cls) if x.name in data}


@dataclass
class CallyProject:
    pass


@dataclass
class CallyService(CallyProject):
    name: str
    environment: str


@dataclass
class CallyBackend:
    config: dict = field(default_factory=dict)
    type: str = 'LocalBackend'
    path: str = 'state/{environment}/{name}'
    path_key: str = 'path'

    def backend_config(self, service: dict) -> dict:
        return {self.path_key: self.path.format(**service), **self.config}


@dataclass
class CallyStackService(CallyService):
    """This dataclass is automatically passed to the CallyStack class during
    instantiation. Allowing access to any service property, though it is
    recommended to use the documented functions.
    """

    stack_type: str
    backend: CallyBackend = field(default_factory=CallyBackend)
    providers: dict = field(default_factory=dict)
    stack_vars: dict = field(default_factory=dict)

    def __setattr__(self, prop, val):
        if prop == 'backend' and isinstance(val, dict):
            super().__setattr__(prop, CallyBackend(**val))
            return
        super().__setattr__(prop, val)

    @property
    def backend_type(self) -> str:
        return self.backend.type

    @property
    def backend_config(self) -> dict:
        return self.backend.backend_config(self.to_dict())

    def get_provider_vars(self, provider: str) -> dict:
        return self.providers.get(provider, {})

    def get_stack_var(self, var: str, default: Optional[Any] = None) -> Any:
        """Any stack var, configured on the service is available via this method, with
        the ability to provide a default.
        """
        return self.stack_vars.get(var, default)

    def to_dict(self) -> dict:
        return asdict(self)


@dataclass
class CallyEnvironment(CallyProject):
    environment: str



import click

from . import CallyCommandClass, CallyConfig, config_types, mixins
from .options import CALLY_ENVIRONMENT_OPTIONS


class CallyEnvironmentConfig(
    CallyConfig[config_types.CallyEnvironment], mixins.CallyEnvironment
):
    CALLY_TYPE = config_types.CallyEnvironment

    @property
    def _settings_kwargs(self) -> dict:
        return {
            'loaders': ['cally.cli.config.loaders.environment'],
            'cally_env': self.environment,
        }


class CallyEnvironmentConfigContext(click.Context):
    def __init__(self, *args, **kwargs) -> None:
        if kwargs.get('obj', None) is None:
            kwargs.update(obj=CallyEnvironmentConfig())
        super().__init__(*args, **kwargs)


def CallyEnvironmentCommand():  # noqa: N802
    class CallyEnvironmentCommandClass(CallyCommandClass):
        context_class = CallyEnvironmentConfigContext
        default_cally_options = CALLY_ENVIRONMENT_OPTIONS

    return CallyEnvironmentCommandClass



PROVIDER_PYPROJECT = """[build-system]
requires = ["setuptools"]
build-backend = "setuptools.build_meta"

[project]
name = "CallyProviders$title"
version = "$version"
requires-python = ">=3.8"
authors = [
  {name = "Cally Generated"},
]
description = "Cally Generated $title Provider"

[tool.setuptools.packages.find]
include = ["cally.*"]

[tool.setuptools.package-data]
"*" = [
    "py.typed",
    "*.tgz",
]"""

PROVIDER_CDKTF = """{
  "language": "python",
  "app": "python3 ./main.py",
  "sendCrashReports": "false",
  "terraformProviders": [
    "$provider@~>$version"
  ],
  "terraformModules": [],
  "codeMakerOutput": "$path",
  "context": {}
}"""



from pathlib import Path
from typing import Union

import click


def ctx_callback(
    ctx: click.Context, param: click.Parameter, value: Union[str, int]
) -> Union[str, int]:
    setattr(ctx.obj, str(param.name), value)
    return value


CALLY_CONFIG_OPTIONS = [
    click.Option(
        ['--config-file'],
        type=click.Path(path_type=Path),
        default=Path(Path.cwd(), 'cally.yaml'),
        envvar='CALLY_CONFIG',
        help='Path to the project config file',
        expose_value=False,
        callback=ctx_callback,
    )
]

CALLY_ENVIRONMENT_OPTIONS = [
    *CALLY_CONFIG_OPTIONS,
    click.Option(
        ['--environment'],
        envvar='CALLY_ENVIRONMENT',
        expose_value=False,
        required=True,
        help='Environment to operate within',
        callback=ctx_callback,
    ),
]

CALLY_SERVICE_OPTIONS = [
    *CALLY_ENVIRONMENT_OPTIONS,
    click.Option(
        ['--service'],
        envvar='CALLY_SERVICE',
        expose_value=False,
        required=True,
        help='Service name to retrieve config details',
        callback=ctx_callback,
    ),
]
